:- ensure_loaded(sq_clpfd).
:- ensure_loaded(sq_prolog).

:- use_module(library(timeout)).
:- use_module(library(clpfd)).

time_out_sec(Sec) :-
	bb_get(time_out_sec, Sec).

solutions(Sols) :-
	bb_get(solutions, Sols).

table_style(Tbl) :-
	bb_get(table_style, Tbl).

set_params(Sec, Sols, Tbl) :-
	bb_put(time_out_sec, Sec),
	bb_put(solutions, Sols),
	bb_put(table_style, Tbl).

tex_file_name(FName) :-
	solutions(Kind),
	atom_concat(table_, Kind, F0),
	atom_concat(F0, '.tbl', FName).

time(Id, Goal, Result-T/B) :-
        time_out_msec(TimeOut),
        init_backtracks(Id),
        statistics(runtime,_),
        time_out(Goal, TimeOut, Result), !,
        statistics(runtime, [_, T]),
        get_backtracks(Id, BType, B),
        show_result(Id, Result, T, BType, B).
time(Id, _Goal, failure-T/B):- 
        statistics(runtime, [_, T]), 
        get_backtracks(Id, BType, B),
        show_result(Id, failure, T, BType, B).

squares(Flags-Over-Lab, Sizes, Limit) :-
        squares_clpfd(Flags-Over-Lab, Sizes, Limit, _Xs, _Ys), solutions(single), !.
squares(prolog, Sizes, Limit) :-
        squares_prolog(Sizes, Limit, _Xs, _Ys), solutions(single), !.
squares(_, _, _).

init_backtracks(_:prolog) :-
        !, bb_put(count, 0).
init_backtracks(_) :-
        fd_statistics(backtracks, _).

get_backtracks(_:prolog, pl, B) :-
        !, bb_get(count, B).
get_backtracks(_, fd, B) :-
        fd_statistics(backtracks,B).

show_result(Limit:Id, Result, T, BType, B) :-
        format(user_error, 
         '~t~d~3+:~w,~30+~w~10+~t~3d~7+ sec, ~a btrks: ~|~t~d~9+\n',
               [Limit,Id,Result,T,BType,B]),
        flush_output(user_error).
        
time_out_msec(MSec) :-
        time_out_sec(Sec), MSec is Sec*1000.

sq(Id, Limit, Result) :-
        sqd(Limit, Sizes),
        time(Limit:Id, squares(Id, Sizes, Limit), Result).

% run all squares problems for all data
run_all(Sec, Sols, Tbl) :-
	set_params(Sec, Sols, Tbl),
	tex_file_name(FName),
	open(FName, append, S), nl(S), nl(S),
	(   test_case(Id),
	    run_one(Id, S),
	    fail
	;   close(S)
	).

run_one(Id, S) :-
	tex_start(S, Id),
	sqd(Limit, _), sq(Id, Limit, Result-TimeBtrk),
	(   Result = success -> tex_column(S, TimeBtrk), fail
	;   true
	), !,
	(   sqd(L, _), L >= Limit, tex_empty(S), fail
	;   tex_end(S)
	).
run_one(_, S) :-
	tex_end(S).

col_delimiters(tbl, '|', ' ').
col_delimiters(tex, '&', '&').

line_delimiter(tbl, '').
line_delimiter(tex, '\\\\').

tex_start(S, Id) :-
	format(S, '~|~w~t~27+', [Id]).

tex_empty(S) :-
	table_style(St),
	col_delimiters(St, Delim1, Delim2),
	format(S, '~|~a~t~a~9+~t~5+', [Delim1,Delim2]).

tex_column(S, Time/B) :-
	table_style(St),
	col_delimiters(St, Delim1, Delim2),
	format(S, '~|~a~t~3d~a~9+', [Delim1,Time,Delim2]),
	(   B >= 100000 -> BK is round(B/1000),
	    format(S, '~|~t~wK~5+', BK)
	;   format(S, '~|~t~w~5+', B)
	).

tex_end(S) :-
	table_style(St),
	line_delimiter(St, LD),
	format(S, '~a~n', [LD]).


sqd(10,  [6,4,4,4,2,2,2,2]).
sqd(20,  [9,8,8,7,5,4,4,4,4,4,3,3,3,2,2,1,1]).
sqd(112, [50,42,37,35,33,29,27,25,24,19,18,17,16,15,11,9,8,7,6,4,2]).
sqd(175, [81,64,56,55,51,43,39,38,35,33,31,30,29,20,18,16,14,9,8,5,4,3,2,1]).
sqd(503, [211,179,167,157,149,143,135,113,100,93,88,87,67,62,50,34,33,27,25,23,22,19,16,15,4]).

% test_case(Params).

test_case(prolog                   ).
			           
test_case([]-spec-[]               ).
			           
test_case([]-card1-[]              ).
test_case([]-card2-[]              ).
test_case([]-ix-[]                 ).

test_case([cap]-ix-[]              ). 
			           
test_case([cap]-ix-[min]           ). 
test_case([cap]-card2-[min]        ). 
test_case([cap]-card1-[min]        ). 
test_case([cap]-spec-[min]         ). 
			           
test_case([cap]-ix-dual            ). 
			           
test_case([cum-e]-ix-[min]         ). 
test_case([cum-e]-ix-dual          ). 
			           
test_case([cum-n]-ix-[min]         ). 
test_case([cum-p]-ix-[min]         ).
test_case([cum-s]-ix-[min]         ).
test_case([cum-d]-ix-[min]         ). 
			           
test_case([cum-ep]-ix-[min]        ).
test_case([cum-es]-ix-[min]        ).
test_case([cum-ed]-ix-[min]        ). 
			           
test_case([dis-n]-none-[min]       ). 
test_case([dis-d]-none-[min]       ). 
test_case([dis-gd]-none-[min]      ).

test_case([dis-g]-none-[min]       ).
test_case([dis-g]-none-dual        ). 
			           			           
test_case([dis-g]-ix-[min]         ). 
test_case([dis-g]-ix-dual          ). 
			           
test_case([cap,cum-e]-ix-[min]     ). 
test_case([cap,cum-e]-ix-dual      ). 

test_case([cap,dis-g]-none-[min]   ).
test_case([cap,dis-g]-none-dual    ).

test_case([cum-e,dis-g]-none-[min] ).
test_case([cum-e,dis-g]-none-dual  ).

			      
%%:- run_all, halt.
